/*
 * Copyright 2009 OW2 Chameleon
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ow2.chameleon.json.impl;

import java.io.IOException;
import java.io.InputStream;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import net.sf.json.JSON;
import net.sf.json.JSONArray;
import net.sf.json.JSONException;
import net.sf.json.JSONNull;
import net.sf.json.JSONObject;
import net.sf.json.JsonConfig;
import net.sf.json.xml.XMLSerializer;

import org.ow2.chameleon.json.JSONBeanService;
import org.ow2.chameleon.json.JSONService;
import org.ow2.chameleon.json.JSONXMLService;

/**
 * JSONService, JSONXMLService and JSONBeanService implementation
 * relying on json-lib.
 * @author <a href="mailto:chameleon-dev@ow2.org">Chameleon Project Team</a>
 */
public class JSONServiceProvider implements JSONService, JSONXMLService,
        JSONBeanService {

    /**
     * Empty JSON Object as a String.
     */
    public static final String EMPTY_JSON = "{}";

    /**
     * Buffer Size for Stream Copy.
     */
    private static final int BUFFER_SIZE = 4 * 1024;


    /**
     * Creates the JSON Object from the given XML.
     * This method uses a {@link XMLSerializer}.
     * @param xml the xml
     * @return the json object (String form)
     * @throws ParseException if the xml cannot be parsed.
     * @see org.ow2.chameleon.json.JSONXMLService#fromXML(java.lang.String)
     */
    public String fromXML(String xml) throws ParseException {
        if (xml == null || xml.length() == 0) {
            return EMPTY_JSON;
        }

        try {
            XMLSerializer serializer = new XMLSerializer();
            JSON json = serializer.read(xml);
            return json.toString();
        } catch (JSONException e) {
            throw new ParseException(e.getMessage(), 0);
        }
    }

    /**
     * Creates the JSON Object from the given XML.
     * This method uses a {@link XMLSerializer}.
     * @param is the xml (input stream)
     * @return the json object (String form)
     * @throws ParseException if the xml cannot be parsed.
     * @throws IOException if the stream cannot be read.
     * @see org.ow2.chameleon.json.JSONXMLService#fromXML(java.lang.String)
     */
    public String fromXML(InputStream is) throws ParseException, IOException {
        String json = read(is);
        return fromXML(json);
    }

    /**
     * Transforms a JSONArray into a List.
     * This list will contain primitive types, map or other
     * list.
     * @param array the array
     * @return the resulting list
     * @throws ParseException if the array cannot be transformed
     */
    private List arrayToList(JSONArray array) throws ParseException {
        if (array == null || array.isEmpty()) {
            return new ArrayList();
        }

        List list = new ArrayList();
        for (int i = 0; i < array.size(); i++) {
            Object object = array.get(i);
            if (object instanceof JSONObject) {
                list.add(fromJSON(object.toString()));
            } else if (object instanceof JSONArray) {
                list.add(arrayToList((JSONArray) object));
            } else {
                list.add(object); // Call string to Value here ?
            }
        }
        return list;

    }

    /**
     * Creates a map representing the JSON Object.
     * The map contains either String, 'primitives' types (object),
     * others Map or List.
     * @param json the JSON Object
     * @return the Map representing the JSON object
     * @throws ParseException if the given string cannot be parsed.
     * @see org.ow2.chameleon.json.JSONService#fromJSON(java.lang.String)
     */
    public Map fromJSON(String json) throws ParseException {

        // Empty case.
        if (json == null || json.length() == 0) {
            return new HashMap();
        }

        try {
            JsonConfig config = new JsonConfig();
            // Disable the Jettison optimization.
            config.setHandleJettisonEmptyElement(false);
            config.setHandleJettisonSingleElementArray(false);
            JSONObject object = JSONObject.fromObject(json, config);
            Map map = new HashMap();
            Iterator iterator = object.keys();
            while (iterator.hasNext()) {
                String key = (String) iterator.next();
                Object obj = object.get(key);

                if (obj instanceof JSONArray) {
                    map.put(key, arrayToList((JSONArray) obj));
                } else if (obj instanceof JSONObject) {
                    // Transfert to map
                    Map inner = fromJSON(obj.toString());
                    map.put(key, inner);

                } else { // Simple type.
                    map.put(key, obj);
                }
            }
            return map;
        } catch (JSONException e) {
            throw new ParseException(e.getMessage(), 0);
        }
    }

    /**
     * Reads an input stream and create the Map representing the
     * read JSON object.
     * The map contains either String, 'primitives' types (object),
     * others Map or List.
     * @param is the JSON Object
     * @return the Map representing the JSON object
     * @throws ParseException if the given string cannot be parsed.
     * @see org.ow2.chameleon.json.JSONService#fromJSON(java.lang.String)
     * @throws IOException if the stream cannot be read.
     * @see org.ow2.chameleon.json.JSONService#fromJSON(java.io.InputStream)
     */
    public Map fromJSON(InputStream is) throws ParseException, IOException {
        if (is == null) {
            return new HashMap();
        } else {
            String json = read(is);
            return fromJSON(json);
        }
    }

    /**
     * Gets the object representing Undefined.
     * @return the object representing Undefined
     * @see org.ow2.chameleon.json.JSONService#getNull()
     */
    public Object getNull() {
        return JSONNull.getInstance();
    }

    /**
     * Reads an input stream and returns the result
     * as a String.
     * @param in the input stream
     * @return the read String
     * @throws IOException if the stream cannot be read.
     */
    private static String read(InputStream in) throws IOException {
        StringBuffer out = new StringBuffer();
        byte[] b = new byte[BUFFER_SIZE];
        for (int n; (n = in.read(b)) != -1;) {
            out.append(new String(b, 0, n));
        }
        return out.toString();
    }

    /**
     * Computes and returns the String representing in
     * JSON-syntax the given map.
     * @param map the map
     * @return the json String
     * @see org.ow2.chameleon.json.JSONService#toJSON(java.util.Map)
     */
    public String toJSON(Map map) {
        if (map == null) {
            return EMPTY_JSON; // Empty object.
        }
        JSONObject json = JSONObject.fromObject(map);
        return json.toString();
    }

    /**
     * Computes and returns the String representing in
     * JSON-syntax the given bean.
     * @param bean a bean
     * @return the json String
     * @see org.ow2.chameleon.json.JSONBeanService#toJSON(java.lang.Object)
     */
    public String toJSON(Object bean) {
        if (bean == null) {
            return EMPTY_JSON;
        }
        JSONObject jsonObject = JSONObject.fromObject(bean);
        return jsonObject.toString();
    }

    /**
     * Instantiates a bean (from the given class) to represent the
     * given JSON object.
     * @param json the json object
     * @param clazz the bean clazz
     * @return the instantiated bean.
     * @throws ParseException if json cannot be read.
     * @see org.ow2.chameleon.json.JSONBeanService#toBean(java.lang.String,
     *  java.lang.Class)
     */
    public Object toBean(String json, Class clazz) throws ParseException {
        final ClassLoader original = Thread.currentThread().getContextClassLoader();
        try {
            // Switch the TCCL to the clazz's classloader
            Thread.currentThread().setContextClassLoader(clazz.getClassLoader());
            JSONObject jsonObject = JSONObject.fromObject(json);
            return JSONObject.toBean(jsonObject, clazz);
        } finally {
            Thread.currentThread().setContextClassLoader(original);
        }

    }

    /**
     * Instantiates a bean (from the given class) to represent the
     * given JSON object.
     * @param json the json object
     * @param clazz the bean clazz
     * @param mapping the mapping for nested beans
     * @return the instantiated bean.
     * @throws ParseException if json cannot be read.
     * @see org.ow2.chameleon.json.JSONBeanService#toBean(java.lang.String,
     *  java.lang.Class)
     *  @since 0.2.2
     */
    public Object toBean(String json, Class clazz, Map mapping) throws ParseException {
        final ClassLoader original = Thread.currentThread().getContextClassLoader();
        try {
            // Switch the TCCL to the clazz's classloader
            Thread.currentThread().setContextClassLoader(clazz.getClassLoader());
            JSONObject jsonObject = JSONObject.fromObject(json);
            return JSONObject.toBean(jsonObject, clazz, mapping);
        } finally {
            Thread.currentThread().setContextClassLoader(original);
        }
    }

    /**
     * Instantiates a bean (from the given class) to represent the
     * given JSON object.
     * @param json the map representing the json object
     * @param clazz the bean clazz
     * @return the instantiated bean.
     * @see org.ow2.chameleon.json.JSONBeanService#toBean(java.util.Map,
     *  java.lang.Class)
     */
    public Object toBean(Map json, Class clazz) {
        final ClassLoader original = Thread.currentThread().getContextClassLoader();
        try {
            // Switch the TCCL to the clazz's classloader
            Thread.currentThread().setContextClassLoader(clazz.getClassLoader());
            JSONObject jsonObject = JSONObject.fromObject(json);
            return JSONObject.toBean(jsonObject, clazz);
        } finally {
            Thread.currentThread().setContextClassLoader(original);
        }
    }
}
